package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kstarbound.RegistryObject
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import java.io.DataInputStream

enum class TileColor {
	DEFAULT,
	RED,
	BLUE,
	GREEN,
	YELLOW,
	BROWN,
	PURPLE,
	BLACK,
	WHITE;

	companion object {
		private val values = values()

		fun of(index: Int): TileColor {
			return values.getOrElse(index) { DEFAULT }
		}
	}
}

data class TileState(
	val material: TileDefinition? = null,
	val modifier: MaterialModifier? = null,
	val color: TileColor = TileColor.DEFAULT,
	val hueShift: Float = 0f,
	val modifierHueShift: Float = 0f,
) {
	private fun with(
		material: TileDefinition? = this.material,
		modifier: MaterialModifier? = this.modifier,
		color: TileColor = this.color,
		hueShift: Float = this.hueShift,
		modifierHueShift: Float = this.modifierHueShift,
	): TileState {
		if (this.material === material && this.modifier === modifier && this.color == color && this.hueShift == hueShift && this.modifierHueShift == modifierHueShift) {
			return this
		} else {
			return TileState(material, modifier, color, hueShift, modifierHueShift)
		}
	}

	fun withHueShift(hueShift: Int): TileState {
		if (hueShift < 0) {
			return with(hueShift = 0f)
		} else if (hueShift >= 255) {
			return with(hueShift = 360f)
		} else {
			return with(hueShift = hueShift / 255f * 360f)
		}
	}

	fun withModifierHueShift(hueShift: Int): TileState {
		if (hueShift < 0) {
			return with(modifierHueShift = 0f)
		} else if (hueShift >= 255) {
			return with(modifierHueShift = 360f)
		} else {
			return with(modifierHueShift = hueShift / 255f * 360f)
		}
	}

	companion object {
		val EMPTY = TileState()

		fun read(
			materialAccess: (Int) -> RegistryObject<TileDefinition>?,
			modifierAccess: (Int) -> RegistryObject<MaterialModifier>?,
			stream: DataInputStream
		): TileState {
			val mat = materialAccess(stream.readUnsignedShort())
			val hue = stream.read()
			val color = stream.read()
			val mod = modifierAccess(stream.readUnsignedShort())
			val modHue = stream.read()

			return TileState(mat?.value, mod?.value, TileColor.of(color), hue / 255f * 360f, modHue / 255f * 360f)
		}
	}
}
